#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import copy
import typing
from typing import Any, Mapping

OPTIONS_STR = "$options"


DEFAULT_MODEL_TYPES: Mapping[str, str] = {
    # CompositeErrorHandler
    "CompositeErrorHandler.error_handlers": "DefaultErrorHandler",
    # CursorPagination
    "CursorPagination.decoder": "JsonDecoder",
    # DatetimeStreamSlicer
    "DatetimeStreamSlicer.end_datetime": "MinMaxDatetime",
    "DatetimeStreamSlicer.end_time_option": "RequestOption",
    "DatetimeStreamSlicer.start_datetime": "MinMaxDatetime",
    "DatetimeStreamSlicer.start_time_option": "RequestOption",
    # DeclarativeSource
    "DeclarativeSource.check": "CheckStream",
    "DeclarativeSource.spec": "Spec",
    "DeclarativeSource.streams": "DeclarativeStream",
    # DeclarativeStream
    "DeclarativeStream.retriever": "SimpleRetriever",
    "DeclarativeStream.schema_loader": "JsonFileSchemaLoader",
    # DefaultErrorHandler
    "DefaultErrorHandler.response_filters": "HttpResponseFilter",
    # DefaultPaginator
    "DefaultPaginator.decoder": "JsonDecoder",
    "DefaultPaginator.page_size_option": "RequestOption",
    "DefaultPaginator.page_token_option": "RequestOption",
    # DpathExtractor
    "DpathExtractor.decoder": "JsonDecoder",
    # HttpRequester
    "HttpRequester.error_handler": "DefaultErrorHandler",
    "HttpRequester.request_options_provider": "InterpolatedRequestOptionsProvider",
    # ListStreamSlicer
    "ListStreamSlicer.request_option": "RequestOption",
    # ParentStreamConfig
    "ParentStreamConfig.request_option": "RequestOption",
    "ParentStreamConfig.stream": "DeclarativeStream",
    # RecordSelector
    "RecordSelector.extractor": "DpathExtractor",
    "RecordSelector.record_filter": "RecordFilter",
    # SimpleRetriever
    "SimpleRetriever.paginator": "NoPagination",
    "SimpleRetriever.record_selector": "RecordSelector",
    "SimpleRetriever.requester": "HttpRequester",
    "SimpleRetriever.stream_slicer": "SingleSlice",
    # SubstreamSlicer
    "SubstreamSlicer.parent_stream_configs": "ParentStreamConfig",
    # AddFields
    "AddFields.fields": "AddedFieldDefinition",
    # CustomStreamSlicer
    "CustomStreamSlicer.end_datetime": "MinMaxDatetime",
    "CustomStreamSlicer.end_time_option": "RequestOption",
    "CustomStreamSlicer.parent_stream_configs": "ParentStreamConfig",
    "CustomStreamSlicer.start_datetime": "MinMaxDatetime",
    "CustomStreamSlicer.start_time_option": "RequestOption",
}

# We retain a separate registry for custom components to automatically insert the type if it is missing. This is intended to
# be a short term fix because once we have migrated, then type and class_name should be requirements for all custom components.
CUSTOM_COMPONENTS_MAPPING: Mapping[str, str] = {
    "CartesianProductStreamSlicer.stream_slicers": "CustomStreamSlicer",
    "CompositeErrorHandler.backoff_strategies": "CustomBackoffStrategy",
    "DeclarativeStream.retriever": "CustomRetriever",
    "DeclarativeStream.transformations": "CustomTransformation",
    "DefaultErrorHandler.backoff_strategies": "CustomBackoffStrategy",
    "DefaultPaginator.pagination_strategy": "CustomPaginationStrategy",
    "HttpRequester.authenticator": "CustomAuthenticator",
    "HttpRequester.error_handler": "CustomErrorHandler",
    "RecordSelector.extractor": "CustomRecordExtractor",
    "SimpleRetriever.stream_slicer": "CustomStreamSlicer",
}


class ManifestComponentTransformer:
    def propagate_types_and_options(
        self, parent_field_identifier: str, declarative_component: Mapping[str, Any], parent_options: Mapping[str, Any]
    ) -> Mapping[str, Any]:
        """
        Recursively transforms the specified declarative component and subcomponents to propagate options and insert the
        default component type if it was not already present. The resulting transformed components are a deep copy of the input
        components, not an in-place transformation.

        :param declarative_component: The current component that is having type and options added
        :param parent_field_identifier: The name of the field of the current component coming from the parent component
        :param parent_options: The options set on parent components defined before the current component
        :return: A deep copy of the transformed component with types and options persisted to it
        """
        propagated_component = dict(copy.deepcopy(declarative_component))
        if "type" not in propagated_component:
            # If the component has class_name we assume that this is a reference to a custom component. This is a slight change to
            # existing behavior because we originally allowed for either class or type to be specified. After the pydantic migration,
            # class_name will only be a valid field on custom components and this change reflects that. I checked, and we currently
            # have no low-code connectors that use class_name except for custom components.
            if "class_name" in propagated_component:
                found_type = CUSTOM_COMPONENTS_MAPPING.get(parent_field_identifier)
            else:
                found_type = DEFAULT_MODEL_TYPES.get(parent_field_identifier)
            if found_type:
                propagated_component["type"] = found_type

        # When there is no resolved type, we're not processing a component (likely a regular object) and don't need to propagate options
        if "type" not in propagated_component:
            return propagated_component

        # Combines options defined at the current level with options from parent components. Options at the current level take precedence
        current_options = dict(copy.deepcopy(parent_options))
        component_options = propagated_component.pop(OPTIONS_STR, {})
        current_options = {**current_options, **component_options}

        # Options should be applied to the current component fields with the existing field taking precedence over options if both exist
        for option_key, option_value in current_options.items():
            propagated_component[option_key] = propagated_component.get(option_key) or option_value

        for field_name, field_value in propagated_component.items():
            if isinstance(field_value, dict):
                # We exclude propagating an option that matches the current field name because that would result in an infinite cycle
                excluded_option = current_options.pop(field_name, None)
                parent_type_field_identifier = f"{propagated_component.get('type')}.{field_name}"
                propagated_component[field_name] = self.propagate_types_and_options(
                    parent_type_field_identifier, field_value, current_options
                )
                if excluded_option:
                    current_options[field_name] = excluded_option
            elif isinstance(field_value, typing.List):
                # We exclude propagating an option that matches the current field name because that would result in an infinite cycle
                excluded_option = current_options.pop(field_name, None)
                for i, element in enumerate(field_value):
                    if isinstance(element, dict):
                        parent_type_field_identifier = f"{propagated_component.get('type')}.{field_name}"
                        field_value[i] = self.propagate_types_and_options(parent_type_field_identifier, element, current_options)
                if excluded_option:
                    current_options[field_name] = excluded_option

        if current_options:
            propagated_component[OPTIONS_STR] = current_options
        return propagated_component



#
# Copyright (c) 2022 Airbyte, Inc., all rights reserved.
#

import re
from typing import Any, Mapping, Set, Tuple, Union

from airbyte_cdk.sources.declarative.parsers.custom_exceptions import CircularReferenceException, UndefinedReferenceException


class ManifestReferenceResolver:
    """
    An incoming manifest can contain references to values previously defined.
    This parser will dereference these values to produce a complete ConnectionDefinition.

    References can be defined using a *ref(<arg>) string.
    ```
    key: 1234
    reference: "*ref(key)"
    ```
    will produce the following definition:
    ```
    key: 1234
    reference: 1234
    ```
    This also works with objects:
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs: "*ref(key_value_pairs)"
    ```
    will produce the following definition:
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs:
      k1: v1
      k2: v2
    ```

    The $ref keyword can be used to refer to an object and enhance it with addition key-value pairs
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs:
      $ref: "*ref(key_value_pairs)"
      k3: v3
    ```
    will produce the following definition:
    ```
    key_value_pairs:
      k1: v1
      k2: v2
    same_key_value_pairs:
      k1: v1
      k2: v2
      k3: v3
    ```

    References can also point to nested values.
    Nested references are ambiguous because one could define a key containing with `.`
    in this example, we want to refer to the limit key in the dict object:
    ```
    dict:
        limit: 50
    limit_ref: "*ref(dict.limit)"
    ```
    will produce the following definition:
    ```
    dict
        limit: 50
    limit-ref: 50
    ```

    whereas here we want to access the `nested.path` value.
    ```
    nested:
        path: "first one"
    nested.path: "uh oh"
    value: "ref(nested.path)
    ```
    will produce the following definition:
    ```
    nested:
        path: "first one"
    nested.path: "uh oh"
    value: "uh oh"
    ```

    to resolve the ambiguity, we try looking for the reference key at the top level, and then traverse the structs downward
    until we find a key with the given path, or until there is nothing to traverse.
    """

    ref_tag = "$ref"

    def preprocess_manifest(self, manifest: Mapping[str, Any]) -> Mapping[str, Any]:
        """
        :param manifest: incoming manifest that could have references to previously defined components
        :return:
        """
        return self._evaluate_node(manifest, manifest, set())

    def _evaluate_node(self, node: Any, manifest: Mapping[str, Any], visited: Set):
        if isinstance(node, dict):
            evaluated_dict = {k: self._evaluate_node(v, manifest, visited) for k, v in node.items() if not self._is_ref_key(k)}
            if self.ref_tag in node:
                # The node includes a $ref key, so we splat the referenced value(s) into the evaluated dict
                evaluated_ref = self._evaluate_node(node[self.ref_tag], manifest, visited)
                if not isinstance(evaluated_ref, dict):
                    return evaluated_ref
                else:
                    # The values defined on the component take precedence over the reference values
                    return evaluated_ref | evaluated_dict
            else:
                return evaluated_dict
        elif isinstance(node, list):
            return [self._evaluate_node(v, manifest, visited) for v in node]
        elif isinstance(node, str) and node.startswith("*ref("):
            if node in visited:
                raise CircularReferenceException(node)
            visited.add(node)
            ret = self._evaluate_node(self._lookup_reference_value(node, manifest), manifest, visited)
            visited.remove(node)
            return ret
        else:
            return node

    def _is_ref_key(self, key):
        return key == self.ref_tag

    def _lookup_reference_value(self, reference: str, manifest: Mapping[str, Any]) -> Any:
        path = re.match("\\*ref\\(([^)]+)\\)", reference).groups()[0]
        if not path:
            raise UndefinedReferenceException(path, reference)
        try:
            return self._read_reference_value(path, manifest)
        except (KeyError, IndexError):
            raise UndefinedReferenceException(path, reference)

    @staticmethod
    def _read_reference_value(ref: str, manifest_node: Mapping[str, Any]) -> Any:
        """
        Read the value at the referenced location of the manifest.

        References are ambiguous because one could define a key containing `.`
        In this example, we want to refer to the `limit` key in the `dict` object:
            dict:
                limit: 50
            limit_ref: "*ref(dict.limit)"

        Whereas here we want to access the `nested.path` value.
          nested:
            path: "first one"
          nested.path: "uh oh"
          value: "ref(nested.path)

        To resolve the ambiguity, we try looking for the reference key at the top level, and then traverse the structs downward
        until we find a key with the given path, or until there is nothing to traverse.

        Consider the path foo.bar.baz. To resolve the ambiguity, we first try 'foo.bar.baz' in its entirety as a top-level key. If this
        fails, we try 'foo' as the top-level key, and if this succeeds, pass 'bar.baz' on as the key to be tried at the next level.
        """
        while ref:
            try:
                return manifest_node[ref]
            except (KeyError, TypeError):
                head, ref = _parse_path(ref)
                manifest_node = manifest_node[head]
        return manifest_node


def _parse_path(ref: str) -> Tuple[Union[str, int], str]:
    """
    Return the next path component, together with the rest of the path.

    A path component may be a string key, or an int index.

    >>> _parse_path("foo.bar")
    "foo", "bar"
    >>> _parse_path("foo[7][8].bar")
    "foo", "[7][8].bar"
    >>> _parse_path("[7][8].bar")
    7, "[8].bar"
    >>> _parse_path("[8].bar")
    8, "bar"
    """
    if match := re.match(r"^\[([0-9]+)\]\.?(.*)", ref):
        idx, rest = match.groups()
        result = int(idx), rest
    else:
        result = re.match(r"([^[.]*)\.?(.*)", ref).groups()
    return result
